/*--------------------------------------------------------------------------
 *  Copyright 2007 utgenome.org
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *--------------------------------------------------------------------------*/
//--------------------------------------
// UTGB Common Project
//
// EagyGeneTrackReader.java
// Since: May 22, 2007
//
// $URL: http://svn.utgenome.org/utgb/trunk/utgb/utgb-core/src/main/java/org/utgenome/format/egt/EagyGeneTrackReader.java $ 
// $Author: leo $
//--------------------------------------
package org.utgenome.format.egt;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;

import org.utgenome.UTGBException;
import org.xerial.util.cui.OptionParser;
import org.xerial.util.cui.OptionParserException;
import org.xerial.util.log.Logger;
import org.xerial.util.xml.InvalidXMLException;
import org.xerial.util.xml.XMLGenerator;

/**
 * Eagy Gene Track to XML converter
 * 
 * @author leo
 * 
 */
public class EagyGeneTrackReader
{
    enum Opt {
        HELP, OUTPUT_FILE, USE_STDIN
    }

    private static Logger _logger = Logger.getLogger(EagyGeneTrackReader.class);
    private static OptionParser<Opt> _opt = new OptionParser<Opt>();

    static
    {
        _opt.addOption(Opt.HELP, "h", "help", "display help messages");
        _opt.addOptionWithArgument(Opt.OUTPUT_FILE, "o", "out", "FILE", "output file name");
        _opt.addOption(Opt.USE_STDIN, "i", "stdin", "read an eagy gene track file from the standard input");
    }

    public static void main(String[] args)
    {
        InputStream in = null;
        OutputStream out = null;
        try
        {
            _opt.parse(args);
            if (_opt.isSet(Opt.HELP))
            {
                displayHelpMessage();
                return;
            }

            if (_opt.isSet(Opt.USE_STDIN))
            {
                in = System.in;
            }
            else
            {
                if (_opt.getArgumentLength() > 0)
                {
                    String eagyGeneTrackFile = _opt.getArgument(0);
                    in = new FileInputStream(eagyGeneTrackFile);
                }
                else
                {
                    throw new OptionParserException("no eagy gene track file is given");
                }
            }

            if (_opt.isSet(Opt.OUTPUT_FILE))
            {
                String outFile = _opt.getValue(Opt.OUTPUT_FILE);
                out = new FileOutputStream(outFile);
            }
            else
                out = System.out;

            EagyGeneTrackReader reader = new EagyGeneTrackReader(in, out);
            reader.translate();

        }
        catch (Exception e)
        {
            System.err.println(e);
        }
        finally
        {
            try
            {
                if (in != null && !in.equals(System.in))
                    in.close();

                if (out != null && !out.equals(System.out))
                    out.close();
            }
            catch (IOException e)
            {
                System.err.println(e);
            }
        }

    }

    private static void displayHelpMessage()
    {
        System.out.println("usage: > java -jar EagyGeneTrackReader [option] (egt file)");
        System.out.println(_opt.helpMessage());
    }

    private BufferedWriter _out = null;
    private BufferedReader _in = null;
    private int _lineCount = 0;

    public EagyGeneTrackReader(InputStream in, OutputStream out)
    {
        this._in = new BufferedReader(new InputStreamReader(in));
        this._out = new BufferedWriter(new OutputStreamWriter(out));
    }

    public void translate() throws IOException, UTGBException
    {
        try
        {
            XMLGenerator xout = new XMLGenerator(_out);
            xout.startTag("eagyGeneTrack");

            String line;
            _lineCount = 0;
            while ((line = _in.readLine()) != null)
            {
                _lineCount++;
                EagyGeneLine eg = new EagyGeneLine(line);
                eg.toXML(xout);
            }

            xout.endTag();
            xout.endDocument();
        }
        catch (InvalidXMLException e)
        {
            throw new UTGBException("line=" + _lineCount + " " + e);
        }

    }

}
