/*--------------------------------------------------------------------------
 *  Copyright 2007 utgenome.org
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *--------------------------------------------------------------------------*/
//--------------------------------------
// GenomeBrowser Project
//
// TrackBase.java
// Since: Jun 12, 2007
//
// $URL: http://svn.utgenome.org/utgb/trunk/utgb/utgb-core/src/main/java/org/utgenome/gwt/utgb/client/track/TrackBase.java $ 
// $Author: leo $
//--------------------------------------
package org.utgenome.gwt.utgb.client.track;

import org.utgenome.gwt.utgb.client.track.bean.TrackBean;
import org.utgenome.gwt.utgb.client.util.Properties;
import org.utgenome.gwt.utgb.client.util.Utilities;
import org.utgenome.gwt.utgb.client.util.xml.DOMUtil;
import org.utgenome.gwt.utgb.client.util.xml.XMLAttribute;
import org.utgenome.gwt.utgb.client.util.xml.XMLUtil;
import org.utgenome.gwt.utgb.client.util.xml.XMLWriter;

import com.google.gwt.core.client.GWT;
import com.google.gwt.user.client.Command;
import com.google.gwt.user.client.DeferredCommand;
import com.google.gwt.xml.client.Node;
import com.google.gwt.xml.client.NodeList;

/**
 * {@link TrackBase} is a base class that supports to implement your own
 * {@link Track}s.
 * 
 * @author leo
 * 
 */
public abstract class TrackBase implements Track
{
    protected TrackGroup _trackGroup = null;
    protected TrackFrame _frame = null;
    protected final TrackInfo _trackInfo;
    private boolean _isInitialized = false;

    private TrackBean _loadedState = null;

    public TrackBase(String trackName)
    {
        _trackInfo = new TrackInfo(trackName);
    }

    public TrackBase(TrackInfo trackInfo)
    {
        _trackInfo = trackInfo;
    }

    public void setUp(TrackFrame trackFrame, TrackGroup group)
    {
    // do nothing
    }

    public TrackGroup getTrackGroup()
    {
        assert (_trackGroup != null);
        return _trackGroup;
    }

    public TrackInfo getTrackInfo()
    {
        return _trackInfo;
    }

    public void setTrackGroup(TrackGroup trackGroup)
    {
        this._trackGroup = trackGroup;
        validateTrackEnvironment();
    }

    public TrackFrame getFrame()
    {
        return _frame;
    }

    public void setFrame(TrackFrame frame)
    {
        this._frame = frame;
        validateTrackEnvironment();
    }

    private void validateTrackEnvironment()
    {
        if (_trackGroup != null && _frame != null)
        {
            _trackGroup.setResizeNotification(false);
            setUp(_frame, _trackGroup);
            if (_loadedState != null)
            {
                // restore trackFrame State
                if (_loadedState.getHeight() > TrackFrameState.DEFAULT_MIN_TRACKFRAME_HEIGHT)
                    _frame.resize(_loadedState.getHeight());

                Boolean packed = _loadedState.getPack();
                if (packed != null)
                {
                    _frame.setPacked(packed.booleanValue());
                }
            }
            _trackGroup.setResizeNotification(true);
            _isInitialized = true;
        }
    }

    public int getDefaultWindowHeight()
    {
        return TrackFrameState.DEFAULT_MIN_TRACKFRAME_HEIGHT;
    }

    public int getMinimumWindowHeight()
    {
        return 0;
    }

    public void onChangeTrackGroupProperty(TrackGroupPropertyChange change)
    {
    // do nothing on default
    }

    public void onChangeTrackWindow(TrackWindow newWindow)
    {
    // do nothing on default
    }

    public void draw()
    {
    // do nothing on default
    }

    public void refresh()
    {
        DeferredCommand.addCommand(new Command() {
            public void execute()
            {
                TrackBase.this.draw();
                getFrame().onUpdateTrackWidget();
            }
        });
    }

    public boolean isInitialized()
    {
        return _isInitialized;
    }

    public void onChangeTrackConfig(TrackConfigChange change)
    {
    // do nothing on default
    }

    public TrackConfig getConfig()
    {
        return null;
    }

    public String getName()
    {
        return getTrackInfo().getTrackName();
    }

    public boolean isTrack()
    {
        return true;
    }

    public boolean isTrackGroup()
    {
        return false;
    }

    public String toXML()
    {
        XMLWriter xmlWriter = new XMLWriter();
        toXML(xmlWriter);
        return xmlWriter.toString();
    }

    public void toXML(XMLWriter xmlWriter)
    {
        xmlWriter.start("track", new XMLAttribute().add("className", getClassName()).add("name", getName()).add(
                "height", getWidget().getOffsetHeight()).add("pack", _frame.isPacked()));

        Properties trackProperties = new Properties();
        saveProperties(trackProperties);
        XMLUtil.toXML(trackProperties, xmlWriter);
        xmlWriter.end(); // track
    }

    public void loadXML(Node trackNode)
    {
        TrackBean trackInfo = new TrackBean();
        trackInfo.setName(DOMUtil.getAttributeValue(trackNode, "name", getName()));
        trackInfo.setHeight(DOMUtil.getIntAttributeValue(trackNode, "height", getDefaultWindowHeight()));
        trackInfo.setPack(new Boolean(DOMUtil.getBooleanAttributeValue(trackNode, "pack", true)));
        // load track properties
        Properties trackProperties = new Properties();
        {
            final NodeList propertiesNodeList = DOMUtil.getChildNodeList(trackNode, "property");

            final int SIZE = propertiesNodeList.getLength();
            for (int i = 0; i < SIZE; i++)
            {
                final Node propertyNode = propertiesNodeList.item(i);
                final Node valueNode = propertyNode.getFirstChild();

                final String key = Utilities.getAttributeValue(propertyNode, "key");
                final String value = valueNode != null ? valueNode.getNodeValue() : "";

                trackInfo.putProperty(key, value);
            }
        }
        load(trackInfo);
    }

    public void load(TrackBean trackInfo)
    {
        _loadedState = trackInfo;

        // track name
        getTrackInfo().setTrackName(trackInfo.getName());
        // track frame state
        // set track properties
        restoreProperties(new Properties(trackInfo.getProperty()));
    }

    public String getClassName()
    {
        return GWT.getTypeName(this);
    }

    /**
     * Override this method to save internal state of this track into a
     * {@link Properties} object
     * 
     * @param xmlWriter
     */
    public void saveProperties(Properties saveData)
    {}

    /**
     * Override this method to restore internal state of this track
     * 
     * @param properties
     */
    public void restoreProperties(Properties properties)
    {}

}
