/*--------------------------------------------------------------------------
 *  Copyright 2007 utgenome.org
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *--------------------------------------------------------------------------*/
//--------------------------------------
// GenomeBrowser Project
//
// TrackGroupPropertyImpl.java
// Since: Jun 12, 2007
//
// $URL: http://svn.utgenome.org/utgb/trunk/utgb/utgb-core/src/main/java/org/utgenome/gwt/utgb/client/track/impl/TrackGroupPropertyImpl.java $ 
// $Author: leo $
//--------------------------------------
package org.utgenome.gwt.utgb.client.track.impl;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import org.utgenome.gwt.utgb.client.track.Track;
import org.utgenome.gwt.utgb.client.track.TrackGroup;
import org.utgenome.gwt.utgb.client.track.TrackGroupPropertyChange;
import org.utgenome.gwt.utgb.client.track.TrackGroupProperty;
import org.utgenome.gwt.utgb.client.track.TrackGroupPropertyChangeListener;
import org.utgenome.gwt.utgb.client.track.TrackGroupPropertyWriter;
import org.utgenome.gwt.utgb.client.track.TrackWindow;
import org.utgenome.gwt.utgb.client.util.xml.XMLUtil;
import org.utgenome.gwt.utgb.client.util.xml.XMLWriter;

/**
 * {@link TrackGroupPropertyImpl} is a set of (key, value) pairs shared within a {@link TrackGroup} Instead of directly
 * access {@link TrackGroupPropertyImpl}, you must read and write values through the {@link TrackGroupProperty} and
 * {@link TrackGroupPropertyWriter}, respectively. This indirection is necessary for broadcasting
 * {@link Track#onChangeTrackGroupProperty(TrackGroupPropertyChange)} and {@link Track#onChangeTrackWindow(TrackWindow)}
 * events to all tracks in a group.
 * 
 * @author leo
 * 
 */
public class TrackGroupPropertyImpl implements TrackGroupProperty, TrackGroupPropertyWriter {
	private final TrackGroup _trackGroup;
	private TrackWindowImpl _trackWindow = new TrackWindowImpl();
	private HashMap _properties = new HashMap();
	private ArrayList _changeListener = new ArrayList();
	

	public TrackGroupPropertyImpl(TrackGroup trackGroup) {
		this._trackGroup = trackGroup;
	}

	public String getProperty(String key) {
		Object value = _properties.get(key);
		if (value != null)
			return value.toString();
		else
			return null;
	}

	public String getProperty(String key, String defaultValue) {
		Object value = _properties.get(key);
		return (value != null) ? value.toString() : defaultValue;
	}

	public TrackWindow getTrackWindow() {
		return _trackWindow;
	}

	public void clear()
	{
	    _properties.clear();
	    _changeListener.clear();
	}
	
	/**
	 * Notify the change to the track group and listeners 
	 * @param change
	 */
	protected void notifyTheChange(TrackGroupPropertyChange change)
	{
		_trackGroup.setResizeNotification(false);
		
		_trackGroup.broadcastPropertyChange(change);
		for(Iterator it = _changeListener.iterator(); it.hasNext(); )
		{
			TrackGroupPropertyChangeListener listener = (TrackGroupPropertyChangeListener) it.next();
			listener.onChangeTrackGroupProperty(change);
		}
		
		_trackGroup.setResizeNotification(true);
		_trackGroup.notifyResize();
	}
	
	public void notifyTrackWindowChange()
	{
		_trackGroup.setResizeNotification(false);
		
		_trackGroup.broadcastTrackWindowChange(_trackWindow);
		
		_trackGroup.setResizeNotification(true);
		_trackGroup.notifyResize();
	}
	
	public void setProperty(String key, String value) {
		_properties.put(key, value);
		
		// notify the change
		notifyTheChange(new TrackPropertyChangeImpl(this, key)); 
	}

	public void setProperty(Map property) {
		if (property == null || property.size() == 0)
			return;
		_properties.putAll(property);
		notifyTheChange(new TrackPropertyChangeImpl(this, property.keySet()));
	}

	public void setTrackWindow(int startOnGenome, int endOnGenome) {
		_trackWindow.setStartOnGenome(startOnGenome);
		_trackWindow.setEndOnGenome(endOnGenome);
		notifyTrackWindowChange();
	}

	public void setTrackWindowSize(int windowWidth) {
		_trackWindow.setWindowWidth(windowWidth);
		notifyTrackWindowChange();
	}

	public Set keySet() {
		return _properties.keySet();
	}

	public void setTrackWindow(TrackWindow newWindow) {
		_trackWindow.set(newWindow);
		notifyTrackWindowChange();
	}

	public void apply(TrackGroupPropertyChange change){
		HashMap map = new HashMap();
		for (Iterator it = change.changedKeySet().iterator(); it.hasNext();) {
			String key = (String) it.next();
			String value = change.getProperty(key);
			map.put(key, value);
		}
		this.setProperty(map);
	}

	public void addTrackGroupPropertyChangeListener(TrackGroupPropertyChangeListener listener) {
		_changeListener.add(listener);
	}

	public void removeTrackGroupPropertyChangeListener(TrackGroupPropertyChangeListener listener) {
		_changeListener.remove(listener);
	}

	
	public void toXML(XMLWriter xmlWriter) {
		xmlWriter.start("groupProperties");
		XMLUtil.toXML(_properties, xmlWriter);
		_trackWindow.toXML(xmlWriter);
		xmlWriter.end();	// group-properties
	}

}

