/*--------------------------------------------------------------------------
 *  Copyright 2007 utgenome.org
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *--------------------------------------------------------------------------*/
//--------------------------------------
// UTGB Common Project
//
// WebServiceMasterBase.java
// Since: 2007/03/27
//
// $URL: http://svn.utgenome.org/utgb/trunk/utgb/utgb-core/src/main/java/org/utgenome/gwt/web/server/WebServiceMaster.java $ 
// $Author: leo $
//--------------------------------------
package org.utgenome.gwt.web.server;

import java.io.BufferedInputStream;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.HashMap;
import java.util.Properties;

import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.servlet.ServletContextEvent;
import javax.servlet.ServletContextListener;

import org.utgenome.UTGBException;
import org.xerial.util.log.Logger;

/**
 * A WebServerConfiguration class supports following facilities:
 *  - Loading a configuration file when the server starts up
 *  -- The server configuration file can be specified with
 *  -- a JVM option: -Dconfig.file=(file path) 
 *  -- an environment variable written in META-INF/context.xml file  
 *     e.g.  <Environment name="config.file" type="java.lang.String" value="(config file path)"/>
 *     
 * context.xml is required when deploying web server application as a war (web archive) file to the Tomcat server. 
 * 
 * -How to Setup
 *  - add the following lines in WEB-INF/web.xml file
 *   <listener>    
 *    <listener-class>org.utgenome.web.WebServerConfiguration</listener-class>
 *   </listener>
 * 
 * In order to add your own initialization/disposal process, 
 * extend the WebServerConfiguration class, then
 * override the initialize() and dispose() methods.  
 * 
 * @author leo
 *
 */
public class WebServiceMaster implements ServletContextListener
{
    private static Logger _logger = Logger.getLogger(WebServiceMaster.class);
    
    protected static WebServiceMaster _instance = null;
    private boolean _isInitialized = false;
    private Properties _property = new Properties();
    private HashMap<String, Object> _glovalVariableHolder = new HashMap<String, Object>();

    static 
    {
        if(_instance == null)
            _instance = new WebServiceMaster();
    }
    
    public WebServiceMaster() {}    
    
    public static boolean isInitialized()
    {
        assert(_instance != null);
        return _instance._isInitialized;
    }
    
    public static String getProperty(String key, String defaultValue)
    {
        return _instance._property.getProperty(key, defaultValue);
    }
    public static String getProperty(String key)
    {
        return getProperty(key, "");
    }

    
    public static Object getVariable(String name)
    {
    	return _instance._glovalVariableHolder.get(name);
    }
    
    public static void setVariable(String name, Object value)
    {
    	_instance._glovalVariableHolder.put(name, value);
    }
    
    public void loadConfigFile(String configFile) throws UTGBException
    {
        try
        {
            _property.load(new BufferedInputStream(new FileInputStream(configFile)));
        }
        catch(IOException e)
        {
            throw new UTGBException(e);
        }
    }
    
    /**
     * Override this method to arrange the initialization process of the web server
     * @throws UTGBException
     */
    public void initialize() throws UTGBException
    {}
    
    /**
     * Override this method to arrange the disposal process of the web server
     * @throws UTGBException
     */
    public void dispose() throws UTGBException 
    {}
    
    
    
    /**
     * called once or more when the server started
     * @param configFile
     * @throws UTGBException 
     */
    synchronized protected static void initializeWebServer() throws UTGBException
    {
        _instance.initialize();
        _instance._isInitialized = true;
    }
    
    /**
     * called once or more when the server terminates  
     * @throws UTGBException 
     */
    synchronized protected static void disposeWebServer() throws UTGBException
    {
        _instance.dispose();
        _instance._isInitialized = false;
    }

    
    
    public void contextInitialized(ServletContextEvent event)
    {
        if(WebServiceMaster.isInitialized())
            return;
   
        
        _logger.info("web server initialization");
        //String configFile = getContextProperty("config.file", System.getProperty("config.file"));
        try {
            /*
            if(configFile == null)
            {
                _logger.error("config.file is not set properly. Use java VM argument -Dconfig.file=... or add Environment tag within context.xml file");
            }
            else
            {
                _logger.info("config file: " + configFile);
                _instance.loadConfigFile(configFile);
            }
            */
            WebServiceMaster.initializeWebServer();    
            _logger.info("web server is initialized");
        } catch (UTGBException e) {
            
            _logger.error("web server initialization failed: " + e);
        }
    }
    
    public static String getContextProperty(String key, String defaultValue)
    {
        String value = null;
        try
        {
            // read the environment variable from the JNDI, 
            // the parmeters of which are configured in META-INF/context.xml in the Tomcat server.
            Context initContext = new InitialContext();
            Context envContext = (Context) initContext.lookup("java:comp/env");
            value = (String) envContext.lookup(key);
            System.setProperty(key, value);    
            return value;
        }
        catch (NamingException e)
        {
            _logger.warn("error while reading an environment context: " + key + ". " + e.getMessage());
            _logger.warn("using the default value: " + defaultValue);
            System.setProperty(key, defaultValue);
            return defaultValue;
        }
    }

    public void contextDestroyed(ServletContextEvent event)
    {
        try
        {
            WebServiceMaster.disposeWebServer();
        }
        catch (UTGBException e)
        {
            _logger.error("web server diposal failed:" + e);
        }
    }

}




