/*--------------------------------------------------------------------------
 *  Copyright 2008 utgenome.org
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *--------------------------------------------------------------------------*/
//--------------------------------------
// utgb-shell Project
//
// UTGBShellSubCommand.java
// Since: Jan 9, 2008
//
// $URL$ 
// $Author$
//--------------------------------------
package org.utgenome.shell;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.net.URL;
import java.util.Properties;

import org.utgenome.gwt.utgb.server.config.TrackConfiguration;
import org.xerial.core.XerialException;
import org.xerial.util.FileResource;
import org.xerial.util.StringUtil;
import org.xerial.util.bean.BeanUtil;
import org.xerial.util.log.Logger;
import org.xerial.util.template.Template;

/**
 * A common implementation of the UTGBShell's sub commands
 * 
 * @author leo
 * 
 */
public abstract class UTGBShellSubCommandBase implements UTGBShellSubCommand {

	private static Logger _logger = Logger.getLogger(UTGBShell.class);

	public static final String APP_FOLDER = "app";
	public static final String SRC_FOLDER = "src/main/java";
	public static final String WEBAPP_FOLDER = "src/main/webapp";
	public static final String EXPLODED_WEBAPP_DIR = "target/utgb";

	public abstract String name();

	public abstract void execute(String[] args) throws UTGBShellException;

	public static String loadUsage(String helpFileName) {
		// display help messages
		StringBuilder out = new StringBuilder();
		try {
			BufferedReader reader = FileResource.open(Create.class, helpFileName);
			String line;
			while ((line = reader.readLine()) != null) {
				out.append(line);
				out.append(StringUtil.NEW_LINE);
			}
			return out.toString();
		}
		catch (IOException e) {
			return "";
		}

	}

	public static boolean isInProjectRoot() {
		return new File("config/track-config.xml").exists();
	}

	public static TrackConfiguration loadTrackConfiguration() throws UTGBShellException {

		if (!isInProjectRoot())
			throw new UTGBShellException("not in the project root folder");

		try {
			BufferedReader reader = new BufferedReader(new FileReader("config/track-config.xml"));
			TrackConfiguration config = BeanUtil.createXMLBean(TrackConfiguration.class, reader);
			return config;
		}
		catch (XerialException e) {
			throw new UTGBShellException("syntax error in track-config.xml file: " + e.getMessage());
		}
		catch (IOException e) {
			throw new UTGBShellException("failed to load config/track-config.xml " + e.getMessage());
		}
	}

	public int compareTo(UTGBShellSubCommand o) {
		return name().compareTo(o.name());
	}

	public static String getPath(File f) {
		return f.getPath().replaceAll("\\\\", "/");
	}

	public static void createContextXML(String contextPath, String projectRoot, boolean reloadable) throws UTGBShellException {
		Properties prop = new Properties();
		prop.put("contextPath", contextPath);
		prop.put("projectRoot", projectRoot);
		prop.put("reloadable", reloadable ? "true" : "false");
		createFileFromTemplate(UTGBShellException.class, "template/java/context.xml.template", "src/main/webapp/META-INF/context.xml", prop, true);
	}

	public static void createFileFromTemplate(Class baseClass, String templateFilePath, String targetFile, Properties prop) throws UTGBShellException {
		createFileFromTemplate(baseClass, templateFilePath, targetFile, prop, false);
	}

	public static void createFileFromTemplate(Class baseClass, String templateFilePath, String targetFile, Properties prop, boolean overWrite)
			throws UTGBShellException {
		// create parent directories;
		File outFile = new File(targetFile);
		if (outFile.exists() && !overWrite) {
			_logger.info(getPath(outFile) + " already exists.");
		}
		else {
			try {
				URL url = FileResource.find(baseClass, templateFilePath);
				if (url == null) {
					throw new UTGBShellException("not found TrackStartup.java.template");
				}
				Template template;

				template = new Template(url.openStream());

				String fileContent = template.apply(prop);

				outFile.getParentFile().mkdirs();

				// output action class
				FileWriter writer = new FileWriter(outFile);
				writer.append(fileContent);
				writer.close();
				_logger.info("create a file: " + getPath(new File(targetFile)));
			}
			catch (IOException e) {
				throw new UTGBShellException(e);
			}
		}

	}

}
