//--------------------------------------
// tomcat-server Project
//
// TomcatServer.java
// Since: Sep 12, 2007
//
// $URL: http://www.xerial.org/svn/project/XerialJ/tags/tomcat-server/tomcat-server-1.3.1/src/main/java/org/xerial/tomcat/TomcatServer.java $ 
// $Author: leo $
//--------------------------------------
package org.xerial.tomcat;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.InetAddress;
import java.net.URISyntaxException;
import java.net.URL;
import java.nio.channels.FileChannel;
import java.util.List;

import org.apache.catalina.Context;
import org.apache.catalina.Engine;
import org.apache.catalina.LifecycleException;
import org.apache.catalina.connector.Connector;
import org.apache.catalina.core.StandardHost;
import org.apache.catalina.startup.Embedded;
import org.apache.catalina.startup.HostConfig;
import org.xerial.util.FileResource;
import org.xerial.util.cui.OptionHandler;
import org.xerial.util.cui.OptionParser;
import org.xerial.util.cui.OptionParserException;
import org.xerial.util.io.VirtualFile;
import org.xerial.util.log.Logger;

/**
 * Embedded Tomcat Server
 * 
 * @author leo
 * 
 */
public class TomcatServer
{
    private static Logger _logger = Logger.getLogger(TomcatServer.class);

    private TomcatServerConfiguration configuration = new TomcatServerConfiguration();
    private Embedded embeddedTomcat = null;
    private Engine tomcatEngine = null;
    private StandardHost tomcatHost = null;

    enum Opt {
        HELP, PORT, CATALINA_BASE, CONTEXT_PATH, DOC_BASE
    }

    /**
     * entry point for running Tomcat from CUI
     * 
     * @param args
     */
    public static void main(String[] args)
    {
        final OptionParser<Opt> optionParser = new OptionParser<Opt>();
        final TomcatServerConfiguration config = new TomcatServerConfiguration();

        // set handlers of command line arguments
        optionParser.addOption(Opt.HELP, "h", "help", "display help message", new OptionHandler<Opt>() {
            public void handle(OptionParser<Opt> parser) throws OptionParserException
            {
                throw new OptionParserException("help");
            }
        });
        optionParser.addOptionWithArgument(Opt.PORT, "p", "port", "PORT", "server port number",
                new OptionHandler<Opt>() {
                    public void handle(OptionParser<Opt> parser) throws OptionParserException
                    {
                        config.setPort(parser.getIntValue(Opt.PORT));
                    }
                });
        optionParser.addOptionWithArgument(Opt.CATALINA_BASE, "t", "tomcat_home", "TOMCAT_HOME",
                "set the home directory of the tomcat", new OptionHandler<Opt>() {
                    public void handle(OptionParser<Opt> parser) throws OptionParserException
                    {
                        config.setCatalinaBase(parser.getValue(Opt.CATALINA_BASE));
                    }
                });
        optionParser.addOptionWithArgument(Opt.CONTEXT_PATH, "c", "contextPath", "path", "/path: context path");
        optionParser.addOptionWithArgument(Opt.DOC_BASE, "d", "docBase", "(WAR or docBase)",
                "path to the war file or docBase to deploy");

        try
        {
            optionParser.parse(args);

            // start the server
            final TomcatServer server = new TomcatServer(config);
            server.start();

            // deploy a given war
            if (optionParser.isSet(Opt.CONTEXT_PATH) && optionParser.isSet(Opt.DOC_BASE))
            {
                File docBase = new File(optionParser.getValue(Opt.DOC_BASE));
                server.addContext(optionParser.getValue(Opt.CONTEXT_PATH), docBase.getAbsolutePath());
            }

            Runtime.getRuntime().addShutdownHook(new Thread() {
                public void run()
                {
                    try
                    {
                        server.stop();
                    }
                    catch (TomcatException e)
                    {
                        _logger.error(e);
                    }
                }
            });

            // wait until Ctrl+C terminates the program
            while (true)
            {
                Thread.sleep(10000000000000000L);
            }

        }
        catch (OptionParserException e)
        {
            if (e.getMessage().equals("help"))
            {
                System.out.println("> TomcatServer [option]");
                System.out.println(optionParser.helpMessage());
            }
            else
                System.err.println(e.getMessage());
            return;
        }
        catch (TomcatException e)
        {
            _logger.error(e);
        }
        catch (InterruptedException e)
        {
            _logger.error(e);
        }

    }

    /**
     * Creates a TomcatServer instance with the specified port
     * 
     * @param port
     *            port used by the tomcat
     */
    public TomcatServer(int port)
    {
        configuration.setPort(port);
    }

    /**
     * Configures the tomcat server
     * 
     * @param configuration
     *            configuration parameters
     */
    public TomcatServer(TomcatServerConfiguration configuration)
    {
        this.configuration = configuration;
    }

    /**
     * Starts the tomcat server
     * 
     * @throws TomcatException
     *             when failed to launch tomcat
     */
    public void start() throws TomcatException
    {
        _logger.debug("port: " + configuration.getPort());
        _logger.debug("catalina base: " + configuration.getCatalinaBase());

        try
        {
            prepareScaffold(configuration.getCatalinaBase());
        }
        catch (IOException e)
        {
            throw new TomcatException(e);
        }

        // Setup the logging. catalina.base is referred in the
        // logging.properties file
        System.setProperty("catalina.base", configuration.getCatalinaBase());
        System.setProperty("java.util.logging.manager", "org.apache.juli.ClassLoaderLogManager");
        System.setProperty("java.util.logging.config.file", configuration.getCatalinaBase()
                + "/conf/logging.properties");

        // Create an embedded server
        embeddedTomcat = new Embedded();
        embeddedTomcat.setCatalinaBase(configuration.getCatalinaBase());

        // Create an engine
        tomcatEngine = embeddedTomcat.createEngine();
        tomcatEngine.setName("xerial");
        tomcatEngine.setDefaultHost("localhost");

        // Create a default virtual host
        String appBase = configuration.getCatalinaBase() + "/webapps";
        _logger.debug("appBase: " + appBase);
        tomcatHost = (StandardHost) embeddedTomcat.createHost("localhost", appBase);

        // Hook up a host config to search for and pull in webapps.
        HostConfig hostConfig = new HostConfig();
        tomcatHost.addLifecycleListener(hostConfig);

        // Tell the engine about the host
        tomcatEngine.addChild(tomcatHost);
        tomcatEngine.setDefaultHost(tomcatHost.getName());

        // Tell the embedded manager about the engine
        embeddedTomcat.addEngine(tomcatEngine);

        // Tell the embedded server about the connector
        InetAddress nullAddr = null;
        Connector connector = embeddedTomcat.createConnector(nullAddr, configuration.getPort(), false);
        connector.setEnableLookups(true);
        embeddedTomcat.addConnector(connector);

        // create the ROOT context
        // Context rootContext = embeddedTomcat.createContext("", "ROOT");
        // tomcatHost.addChild(rootContext);

        // // add manager context
        // Context managerContext = embeddedTomcat.createContext("/manager",
        // "manager");
        // managerContext.setPrivileged(true);
        // tomcatHost.addChild(managerContext);

        // start up the tomcat
        try
        {
            embeddedTomcat.start();
        }
        catch (LifecycleException e)
        {
            _logger.error(e);
            String m = e.getMessage();
            if (m != null && m.indexOf("already in use") != -1)
                m = "port " + configuration.getPort()
                        + " is already in use. You probably have another tomcat listening the same port";

            releaseTomcatResources();

            throw new TomcatException(e);
        }

    }

    public void registerWAR(String contextPath, String pathToTheWarFile) throws TomcatException
    {
        addContext(contextPath, pathToTheWarFile);
    }

    public void addContext(String contextPath, String docBase) throws TomcatException
    {
        if (embeddedTomcat == null)
            throw new TomcatException("tomcat server is not started yet.");

        _logger.debug("deploy: contextPath=" + contextPath + ", docBase=" + docBase);

        Context context = embeddedTomcat.createContext(contextPath, docBase);
        // load the META-INF/context.xml
        context.setConfigFile(docBase + "/META-INF/context.xml");
        tomcatHost.addChild(context);
    }

    private void releaseTomcatResources()
    {
        embeddedTomcat = null;
        tomcatEngine = null;
        tomcatHost = null;
    }

    /**
     * Stops the tomcat server
     * 
     * @throws TomcatException
     *             failed to stop tomcat
     */
    public void stop() throws TomcatException
    {
        if (embeddedTomcat != null)
        {
            try
            {
                embeddedTomcat.stop();
            }
            catch (LifecycleException e)
            {
                throw new TomcatException(e);
            }
            finally
            {
                releaseTomcatResources();
            }
        }
    }

    /**
     * Creates the folder structure for the tomcat
     * 
     * @param catalinaBase
     * @throws IOException
     */
    private void prepareScaffold(String catalinaBase) throws IOException
    {
        // create the base folder for the scaffold
        File tomcatBase = new File(catalinaBase);

        List<VirtualFile> tomcatResources = FileResource.listResources("org.xerial.tomcat.scaffold");

        if (tomcatResources.size() <= 0)
            throw new IllegalStateException("org.xerial.tomcat.scaffold is not found");

        // sync scaffoldDir with tomcatBase
        for (VirtualFile vf : tomcatResources)
        {
            String srcLogicalPath = vf.getLogicalPath();
            File targetFile = new File(tomcatBase, srcLogicalPath);

            if (vf.isDirectory())
            {
                targetFile.mkdirs();
            }
            else
            {
                _logger.debug("rsync: src=" + vf.getLogicalPath() + " target=" + targetFile.getPath());

                File parentFolder = targetFile.getParentFile();
                parentFolder.mkdirs();

                // copy the file content without overwrite
                if (!targetFile.exists())
                {
                    InputStream reader = vf.getURL().openStream();
                    FileOutputStream writer = new FileOutputStream(targetFile);
                    byte[] buffer = new byte[1024];
                    int bytesRead = 0;
                    while ((bytesRead = reader.read(buffer)) > 0)
                    {
                        writer.write(buffer, 0, bytesRead);
                    }
                }
            }
        }

    }

    private static String packagePath(String packageName)
    {
        String packageAsPath = packageName.replaceAll("\\.", "/");
        return packageAsPath.endsWith("/") ? packageAsPath : packageAsPath + "/";
    }

    public File findDir(String resourceName)
    {
        String resourcePath = packagePath(resourceName);
        if (!resourcePath.startsWith("/"))
            resourcePath = "/" + resourcePath;
        URL resourceURL = this.getClass().getResource(resourcePath);
        if (resourceURL != null)
        {
            _logger.debug("found resource:" + resourceURL);
            String protocol = resourceURL.getProtocol();
            if (protocol.equals("file"))
            {
                try
                {
                    return new File(resourceURL.toURI());
                }
                catch (URISyntaxException e)
                {
                    _logger.error(e);
                }
            }

            return new File(resourceURL.toString());
        }
        return null;
    }

    /**
     * Sync the srcDir content to the targetDir (no overwrite is performed from
     * the source to the target)
     * 
     * @param srcDir
     * @param targetDir
     * @throws IOException
     */
    private void rsync(File srcDir, File targetDir) throws IOException
    {
        _logger.trace("rsync: src=" + srcDir + ", dest=" + targetDir);
        /*
         * if (!srcDir.isDirectory()) throw new IllegalStateException(srcDir + "
         * is not a directory");
         */

        // omit the .svn folders
        if (srcDir.getName() == ".svn")
            return;

        // create directories
        targetDir.mkdirs();

        for (File file : srcDir.listFiles())
        {
            String fileName = file.getName();
            if (file.isDirectory())
            {
                rsync(file, new File(targetDir, fileName));
                continue;
            }

            if (fileName.startsWith("~"))
                continue;

            File newFile = new File(targetDir, fileName);
            // copy a file
            copyWithNoOverwrite(file, newFile);
        }
    }

    private void copyWithNoOverwrite(File from, File to) throws IOException
    {
        if (to.exists())
            return;
        FileChannel srcChannel = new FileInputStream(from).getChannel();
        FileChannel destChannel = new FileOutputStream(to).getChannel();
        srcChannel.transferTo(0, srcChannel.size(), destChannel);
        srcChannel.close();
        destChannel.close();
    }

}
